#include <assert.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <alloca.h>
#include "bignum.h"
#include "util.h"

#include <math.h>											/* WARRING. Need to link with math.h */

#define BASICSZ 2

static void kmul_recursive(struct bignum *xy, unsigned short *x, unsigned short *y, int n);
static void kmul_basic(struct bignum *xy, unsigned short *x, unsigned short *y, int n);
static void k_add3(struct bignum *z0, struct bignum *z1, struct bignum *z2, int n);

														/* Number of elements in mass. */
#define NELEMS 8
struct bignum **mass;

int nesting;

int
bnum_kmul(struct bignum *c, struct bignum *a, struct bignum *b)
{
	int n;
	assert(c != NULL && b != NULL && a != NULL);

	if (a->top < b->top) {
		struct bignum *tmp;
		tmp = a;
		a = b;
		b = tmp;
	}

	n = a->top + 1;

	bnum_ensure_len(b, a->len);

	int temp, i;
	temp = 1 + log(a->len) / log(2);									/* Need to optimize. */
	temp = 2 * temp * NELEMS;
	mass = malloc(temp * sizeof(*mass));
	for (i = 0; i < temp; i++) {
		mass[i] = bnum_new();
	}

	
	if (n > BASICSZ)
		kmul_recursive(c, a->val, b->val, n);
	else
		kmul_basic(c, a->val, b->val, n);

	c->sign = a->sign ^ b->sign;
	
	for (i = 0; i < temp; i++)
		bnum_free(mass[i]);
	free(mass);
	return 0;
}

static void
kmul_recursive(struct bignum *xy, unsigned short *x, unsigned short *y, int n)
{
	int m;
	struct bignum *z0, *z1, *z2;
	struct bignum *sx, *sy, *bnp1, *bnp2, *bnp3;
	int tmp;

	z0 = mass[NELEMS * nesting];
	z1 = mass[NELEMS * nesting + 1];
	z2 = mass[NELEMS * nesting + 2];
	bnp1 = mass[NELEMS * nesting + 3];
	bnp2 = mass[NELEMS * nesting + 4];
	bnp3 = mass[NELEMS * nesting + 5];
	sx = mass[NELEMS * nesting + 6];
	sy = mass[NELEMS * nesting + 7];
	

	m = n / 2;
	nesting++;
	
	if ( m <= BASICSZ)
		kmul_basic(z0, x, y, m);
	else
		kmul_recursive(z0, x, y, m);

	if (n - m <= BASICSZ)
		kmul_basic(z2, (x + m), (y + m), (n - m));
	else
		kmul_recursive(z2, (x + m), (y + m), (n - m));

	bnum_set_safe(bnp1, x, m, 0);
	bnum_set_safe(bnp2, (x + m), (n - m), 0);
	bnum_add(sx, bnp1,bnp2);								/* sx = x1 + x0. */

	bnum_set_safe(bnp1, y, m, 0);
	bnum_set_safe(bnp2, (y + m), (n - m), 0);
	bnum_add(sy, bnp1,bnp2);								/* sy = y1 + y0. */

	tmp = (sx->top >sy->top) ? sx->top + 1 : sy->top + 1;

	bnum_ensure_len(sx, tmp);
	bnum_ensure_len(sy, tmp);

	if (tmp <= BASICSZ) 
		kmul_basic(z1, sx->val, sy->val, tmp);
	else
		kmul_recursive(z1, sx->val, sy->val, tmp);					/* z1 = (x1 + x0)* (y1 + y0) */

	bnum_sub(bnp3, z1, z0);
	bnum_sub(z1, bnp3, z2);

	k_add3(z0, z1, z2, n);

	bnum_copy(xy, z0);

	nesting--;
}

static void
kmul_basic(struct bignum *xy, unsigned short *x, unsigned short *y, int n)
{
	int m;
	unsigned short z0[2], z2[2];
	unsigned short sx[2], sy[2];

	struct bignum *mp1, *mp2, *res;

	mp1 = mass[NELEMS * nesting];
	mp2 = mass[NELEMS * nesting + 1];
	res = mass[NELEMS * nesting + 2];

	m = n / 2;

	*((int*)z0) = *x * *y;
	if (n == 1) {
		*((int*)z2) = 0x0000;
		*((int*)sx) = *x;
		*((int*)sy) = *y;
	} else {
		*((int*)z2) = *(x+1) * *(y+1);
		*((int*)sx) = *x + *(x+1);
		*((int*)sy) = *y + *(y+1);
	}

	bnum_set_safe(mp1, sx, 2, 0);
	bnum_set_safe(mp2, sy, 2, 0);

	bnum_mul(res, mp1, mp2);

	bnum_set_safe(mp1, z0, 2, 0);

	bnum_sub(mp2, res, mp1);								/* mp2 = sx * sy - z0. */

	bnum_set_safe(mp1, z2, 2, 0);
	bnum_sub(res, mp2, mp1);								/* z2 = res = mp2 - z2. */
	bnum_set_safe(xy, z0, 2, 0);

	k_add3(xy, res, mp1, 2);
}

static void
k_add3(struct bignum *z0, struct bignum *z1, struct bignum *z2, int n)
{
	int i, cc, m, tmp;
	m = n / 2;
	tmp = 2*m;
	cc = 0;
	
	bnum_ensure_len(z0, 2*n);

	for (i = 0; i < 2*n; i++) {
		if (i <= z0->top)
			cc += z0->val[i];

		if (i >= m && (i - m) <= z1->top)
			cc += z1->val[i-m];

		if (i >= tmp  && (i - tmp) <= z2->top)
			cc += z2->val[i-tmp];

		if (cc >= BIGNUM_BASE) {
			z0->val[i] = cc % BIGNUM_BASE;
			cc = cc / BIGNUM_BASE;
		} else {
			z0->val[i] = cc;
			cc = 0;
		}
	}
	z0->top = i-1;

	while ( z0->val[z0->top] == 0) {
		if (z0->top >=0)
			z0->top--;
		else
			break;
	}
}
